import React, { useEffect } from 'react';

function styleInject(css, ref) {
  if ( ref === void 0 ) ref = {};
  var insertAt = ref.insertAt;

  if (!css || typeof document === 'undefined') { return; }

  var head = document.head || document.getElementsByTagName('head')[0];
  var style = document.createElement('style');
  style.type = 'text/css';

  if (insertAt === 'top') {
    if (head.firstChild) {
      head.insertBefore(style, head.firstChild);
    } else {
      head.appendChild(style);
    }
  } else {
    head.appendChild(style);
  }

  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    style.appendChild(document.createTextNode(css));
  }
}

var css_248z$1 = ":root {\n  --background: #fff;\n  --font-color: #494949;\n}\n\n@media (prefers-color-scheme: dark) {\n  :root {\n    --background: #3c3c3c;\n    --font-color: #fafafa;\n  }\n}\n.test-component {\n  background-color: var(--background);\n  color: var(--font-color);\n  border: 1px solid #131111;\n  padding: 16px;\n  width: 360px;\n  text-align: center;\n}\n.test-component .heading {\n  font-family: \"Avenir Next\", Helvetica, Arial, sans-serif;\n  font-size: 40px;\n  font-weight: bold;\n}";
styleInject(css_248z$1);

var TestComponent = function (_a) {
    var heading = _a.heading, content = _a.content;
    return (React.createElement("div", { "data-testid": "test-component", className: "test-component" },
        React.createElement("h1", { "data-testid": "test-component__heading", className: "heading" }, heading),
        React.createElement("div", { "data-testid": "test-component__content" }, content)));
};

/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (_) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

// ### Async & I/O

// Return Promise for getting an image.
// - use: imagePromise('./path/to/img').then(img => imageFcn(img))
/**
 * Return a Promise for getting an image.
 *
 * use: imagePromise('./path/to/img').then(img => imageFcn(img))
 * or: await imagePromise('./path/to/img')
 *
 * @param {URL} url string for path to image
 * @return {Promise} A promise resolving to the image
 */
function imagePromise(url) {
  return new Promise((resolve, reject) => {
    const img = new Image();
    img.crossOrigin = "Anonymous";
    img.onload = () => resolve(img);
    // img.onerror = () => reject(Error(`Could not load image ${url}`))
    img.onerror = () => reject(`Could not load image ${url}`);
    img.src = url;
  });
}

// deprecated, use: await fetch(url).then(res => res.<type>())
// https://developer.mozilla.org/en-US/docs/Web/API/Response#methods
// type = "arrayBuffer" "blob" "formData" "json" "text"
// export async function fetchType(url, type = 'text') {
//     const response = await fetch(url)
//     if (!response.ok) throw Error(`Not found: ${url}`)
//     const value = await response[type]()
//     return value
// }

// // Similar pair for requestAnimationFrame
// export function rafPromise() {
//     return new Promise(resolve => requestAnimationFrame(resolve))
// }
// export async function rafLoop(fcn, steps = -1) {
//     let i = 0
//     while (i++ !== steps) {
//         fcn(i - 1)
//         await rafPromise()
//     }
// }
//

// ### Canvas

// import { inWorker } from './dom.js'

function offscreenOK() {
  // return !!self.OffscreenCanvas
  // return typeof OffscreenCanvas !== 'undefined'
  return inWorker();
}

/**
 * Create a blank 2D canvas of a given width/height.
 *
 * @param {number} width The canvas height in pixels
 * @param {number} height The canvas width in pixels
 * @param {boolean} [offscreen=offscreenOK()] If true, return "Offscreen" canvas
 * @return {Canvas} The resulting Canvas object
 */
function createCanvas(width, height, offscreen = offscreenOK()) {
  if (offscreen) return new OffscreenCanvas(width, height);
  const can = document.createElement("canvas");
  can.width = width;
  can.height = height;
  return can;
}
/**
 * As above, but returing the 2D context object instead of the canvas.
 * Note ctx.canvas is the canvas for the ctx, and can be use as an image.
 *
 * @param {number} width The canvas height in pixels
 * @param {number} height The canvas width in pixels
 * @param {boolean} [offscreen=offscreenOK()] If true, return "Offscreen" canvas
 * @return {Context2D} The resulting Canvas's 2D context
 */
function createCtx(width, height, offscreen = offscreenOK()) {
  const can = createCanvas(width, height, offscreen);
  return can.getContext("2d");
}

// Set the ctx/canvas size if differs from width/height.
// It does not install a transform and assumes there is not one currently installed.
// The World object can do that for AgentSets.
function setCanvasSize(can, width, height) {
  if (can.width !== width || can.height != height) {
    can.width = width;
    can.height = height;
  }
}

// Install identity transform for this context.
// Call ctx.restore() to revert to previous transform.
function setIdentity(ctx) {
  ctx.save(); // NOTE: Does not change state, only saves current state.
  ctx.setTransform(1, 0, 0, 1, 0, 0); // or ctx.resetTransform()
}
// Set the text font, align and baseline drawing parameters.
// Ctx can be either a canvas context or a DOM element
// See [reference](http://goo.gl/AvEAq) for details.
// * font is a HTML/CSS string like: "9px sans-serif"
// * align is left right center start end
// * baseline is top hanging middle alphabetic ideographic bottom
function setTextProperties(
  ctx,
  font,
  textAlign = "center",
  textBaseline = "middle"
) {
  Object.assign(ctx, { font, textAlign, textBaseline });
}

// Draw string of the given color at the xy location, in ctx pixel coords.
// Use setIdentity .. reset if a transform is being used by caller.
function drawText(ctx, string, x, y, color, useIdentity = true) {
  if (useIdentity) setIdentity(ctx);
  ctx.fillStyle = color.css || color; // OK to use Color.typedColor
  ctx.fillText(string, x, y);
  if (useIdentity) ctx.restore();
}

// # Draw string of the given color at the xy location, in ctx pixel coords.
// # Use setIdentity .. reset if a transform is being used by caller.
// ctxDrawText: (ctx, string, x, y, color, setIdentity = true) ->
//   @setIdentity(ctx) if setIdentity
//   ctx.fillStyle = color.css # @colorStr color
//   ctx.fillText(string, x, y)
//   ctx.restore() if setIdentity

// Return the (complete) ImageData object for this context object
function ctxImageData(ctx) {
  return ctx.getImageData(0, 0, ctx.canvas.width, ctx.canvas.height);
}
// Clear this context using the cssColor.
// If no color or if color === 'transparent', clear to transparent.
function clearCtx(ctx, cssColor = undefined) {
  const { width, height } = ctx.canvas;
  setIdentity(ctx);
  if (!cssColor || cssColor === "transparent") {
    ctx.clearRect(0, 0, width, height);
  } else {
    cssColor = cssColor.css || cssColor;
    ctx.fillStyle = cssColor;
    ctx.fillRect(0, 0, width, height);
  }
  ctx.restore();
}
// Fill this context with the given image. Will scale image to fit ctx size.
function fillCtxWithImage(ctx, img) {
  setIdentity(ctx); // set/restore identity
  ctx.drawImage(img, 0, 0, ctx.canvas.width, ctx.canvas.height);
  ctx.restore();
}

function inWorker() {
  return !inNode() && typeof self.window === "undefined";
}

function inNode() {
  return typeof global !== "undefined";
}

// Return random int/float in [0,max) or [min,max) or [-r/2,r/2)
/**
 * Returns an int in [0, max), equal or grater than 0, less than max
 *
 * @param {number} max The max integer to return
 * @return {number} an integer in  [0, max)
 */
const randomInt = (max) => Math.floor(Math.random() * max);

/**
 * Returns an int in [min, max), equal or grater than min, less than max
 *
 * @param {number} min The min integer to return
 * @param {number} max The max integer to return
 * @return {number} an integer in  [min, max)
 */
const randomInt2 = (min, max) =>
  min + Math.floor(Math.random() * (max - min));

/**
 * Returns a random float in [min, max)
 *
 * @param {number} min The min float to return
 * @param {number} max The max float to return
 * @return {number} a float in [min, max)
 */
const randomFloat2 = (min, max) => min + Math.random() * (max - min);

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Operators/Remainder
// The modulus is defined as: x - y * floor(x / y)
// It is not %, the remainder function.
/**
 * A true modulus function, differing from the % remainder operation.
 *
 * @param {number} v The value to calculate the modulus of
 * @param {number} n The number relative to which the modulus is calculated.
 * @returns {number} The value of v mod n
 */
const mod = (v, n) => ((v % n) + n) % n; // v - n * Math.floor(v / n)

/**
 * Clamp a float to be between [min, max).
 *
 * @param {number} v value to clamp between min & max
 * @param {number} min min value
 * @param {number} max max value
 * @return {number} a float between min/max
 */
function clamp(v, min, max) {
  if (v < min) return min;
  if (v > max) return max;
  return v;
}

// Return a linear interpolation between lo and hi.
// Scale is in [0-1], a percentage, and the result is in [lo,hi]
// If lo>hi, scaling is from hi end of range.
// [Why the name `lerp`?](http://goo.gl/QrzMc)
const lerp = (lo, hi, scale) =>
  lo <= hi ? lo + (hi - lo) * scale : lo - (lo - hi) * scale;
// Calculate the lerp scale given lo/hi pair and a number between them.
// Clamps number to be between lo & hi.
function lerpScale(number, lo, hi) {
  if (lo === hi) throw Error("lerpScale: lo === hi");
  number = clamp(number, lo, hi);
  return (number - lo) / (hi - lo);
}
// Return a string representation of an array of arrays
const arraysToString = (arrays) => arrays.map((a) => `${a}`).join(",");

// Execute fcn for all own member of an obj or array (typed OK).
// Return input arrayOrObj, transformed by fcn.
// - Unlike forEach, does not skip undefines.
// - Like map, forEach, etc, fcn = fcn(item, key/index, obj).
// - Alternatives are: `for..of`, array map, reduce, filter etc
function forLoop(arrayOrObj, fcn) {
  if (arrayOrObj.slice) {
    // typed & std arrays
    for (let i = 0, len = arrayOrObj.length; i < len; i++) {
      fcn(arrayOrObj[i], i, arrayOrObj);
    }
  } else {
    // obj
    Object.keys(arrayOrObj).forEach((k) => fcn(arrayOrObj[k], k, arrayOrObj));
  }
  // return arrayOrObj
}
// Repeat function f(i, a) n times, i in 0, n-1
// a is optional array, default a new Array.
// Return a.
/**
 * Repeats the function f(i, a) i in 0, n-1.
 * a is an optional array, default a new empty Array
 * returns a, only needed if f() places data in a
 *
 * @param {number} n An integer number of times to run f()
 * @param {function} f The function called.
 * @param {Array} [a=[]] An optional array for use by f()
 * @return {Array} The result of calling f() n times
 */
function repeat(n, f, a = []) {
  for (let i = 0; i < n; i++) f(i, a);
  return a;
}
// Repeat function n times, incrementing i by step each call.
function step(n, step, f) {
  for (let i = 0; i < n; i += step) f(i);
}

/**
 * Return random item from an array
 *
 * @param {Array} An array to choose from
 * @returns {any} The chosen item
 */
// Return random one of array items.
const oneOf = (array) => array[randomInt(array.length)];

// Random key/val of object
const oneKeyOf = (obj) => oneOf(Object.keys(obj));
const oneValOf = (obj) => obj[oneKeyOf(obj)];

// Return a "ramp" (array of uniformly ascending/descending floats)
// in [start,stop] with numItems (positive integer > 1).
// OK for start>stop. Will always include start/stop w/in float accuracy.
function floatRamp(start, stop, numItems) {
  // NOTE: start + step*i, where step is (stop-start)/(numItems-1),
  // has float accuracy problems, must recalc step each iteration.
  if (numItems <= 1) throw Error("floatRamp: numItems must be > 1");
  const a = [];
  for (let i = 0; i < numItems; i++) {
    a.push(start + (stop - start) * (i / (numItems - 1)));
  }
  return a;
}
// Integer version of floatRamp, start & stop integers, rounding each element.
// Default numItems yields unit step between start & stop.
function integerRamp(
  start,
  stop,
  numItems = Math.abs(stop - start) + 1
) {
  return floatRamp(start, stop, numItems).map((a) => Math.round(a));
}

// // Return a new shallow of array (either Array or TypedArray)
// export function clone(array) {
//     return array.slice(0)
// }

// // Simple uniq on sorted or unsorted array.
// export const uniq = array => Array.from(new Set(array))

// import { isObject, isTypedArray } from './types.js'
// import { forLoop } from './objects.js'

// ### OofA/AofO

function isOofA(data) {
  if (!isObject(data)) return false;
  return Object.values(data).every((v) => isTypedArray(v));
}
function oofaObject(oofa, i, keys) {
  const obj = {};
  keys.forEach((key) => {
    obj[key] = oofa[key][i];
  });
  return obj;
}
function toAofO(oofa, keys = Object.keys(oofa)) {
  const length = oofa[keys[0]].length;
  const aofo = new Array(length);
  forLoop(aofo, (val, i) => {
    aofo[i] = oofaObject(oofa, i, keys);
  });
  return aofo;
}
// ### Types

// Fix the javascript typeof operator https://goo.gl/Efdzk5
const typeOf = (obj) =>
  ({}.toString
    .call(obj)
    .match(/\s(\w+)/)[1]
    .toLowerCase());
const isType = (obj, string) => typeOf(obj) === string;
const isOneOfTypes = (obj, array) => array.includes(typeOf(obj));

const isString = (obj) => isType(obj, "string");
const isObject = (obj) => isType(obj, "object");

// Is a number an integer (rather than a float w/ non-zero fractional part)
const isCanvas = (obj) =>
  isOneOfTypes(obj, ["htmlcanvaselement", "offscreencanvas"]);
const isImageable = (obj) =>
  isOneOfTypes(obj, [
    "image",
    "htmlimageelement",
    "htmlcanvaselement",
    "offscreencanvas",
    "imagebitmap",
  ]);

// Typed Arrays:
const isTypedArray = (obj) => typeOf(obj.buffer) === "arraybuffer";

// Unused:
// export const isWebglArray = obj =>
//     Array.isArray(obj) && obj.length === 3 && util.arrayMax(obj) <= 1
// isHtmlElement: obj => /^html.*element$/.test(typeOf(obj))
// isImage: obj => isType(obj, 'image')
// isImageBitmap: obj => isType(obj, 'imagebitmap')
// // Is undefined, null, bool, number, string, symbol
// isPrimitive: obj => obj == null || 'object' != typeof obj
// Return array's type (Array or TypedArray variant)
// typeName: obj => obj.constructor.name

// class World defines the coordinate system for the model.
// It will be  upgraded with methods converting from other
// transforms like  GIS and DataSets.

// const defaultZ = (maxX, maxY) => Math.max(maxX, maxY)

/**
 * @private
 * @typedef {Object} WorldOptions
 * @property {number} minX Max world patch x integer value
 * @property {number} minY Max world patch y integer value
 * @property {number} minZ Max world patch z integer value
 * @property {number} maxX Min world patch x integer value
 * @property {number} maxY Min world patch y integer value
 * @property {number} maxZ Min world patch z integer value
 */

/**
 * @description
 * Class World defines the coordinate system for the model.
 * It has transforms for multiple coordinate systems.
 *
 * The world is defined by an object with 6 properties:
 *
 *          WorldOptions = {
 *              minX: integer,
 *              maxX: integer,
 *              minY: integer,
 *              maxY: integer,
 *              minZ: integer,
 *              maxZ: integer,
 *          }
 */
class World {
  maxX = 16;
  maxY = 16;
  maxZ = 16;
  minX = -16;
  minY = -16;
  minZ = -16;

  /**
   * Create a new World object given an Object with optional
   * minX, maxX, minY, maxY, minZ, maxZ overriding class properties.
   * @param {World|WorldOptions|Object} options Object with overrides for class properties
   */
  constructor(options = {}) {
    Object.assign(this, options); // set the option override values
    this.setWorld(); // convert these to rest of world parameters
  }

  /**
   * Return a default options object, origin at center.
   *
   * @param {number} [maxX=16] Integer max X value
   * @param {number} [maxY=maxX] Integer max Y value
   * @param {number} [maxZ=Math.max(maxX, maxY)] Integer max Z value
   * @returns {WorldOptions}
   */
  static defaultOptions(maxX = 16, maxY = maxX, maxZ = Math.max(maxX, maxY)) {
    return {
      minX: -maxX,
      maxX: maxX,
      minY: -maxY,
      maxY: maxY,
      minZ: -maxZ,
      maxZ: maxZ,
    };
  }
  /**
   * Factory to create a default World instance.
   *
   * @param {number} [maxX=16] Integer max X value
   * @param {number} [maxY=maxX] Integer max Y value
   * @param {number} [maxZ=Math.max(maxX, maxY)] Integer max Z value
   * @returns {World}
   */
  static defaultWorld(maxX = 16, maxY = maxX, maxZ = maxX) {
    return new World(World.defaultOptions(maxX, maxY, maxZ));
  }

  // ======================

  setWorld() {
    // Complete properties derived from minX/Y, maxX/Y (patchSize === 1)

    let { minX, maxX, minY, maxY, minZ, maxZ } = this;

    forLoop({ minX, maxX, minY, maxY, minZ, maxZ }, (val, key) => {
      if (!Number.isInteger(val))
        throw Error(`${key}:${val} must be an integer`);
    });

    this.numX = this.width = maxX - minX + 1;
    this.numY = this.height = maxY - minY + 1;
    // if (maxZ == null) maxZ = this.maxZ = Math.max(this.width, this.height)
    this.numZ = this.depth = maxZ - minZ + 1;

    this.minXcor = minX - 0.5;
    this.maxXcor = maxX + 0.5;
    this.minYcor = minY - 0.5;
    this.maxYcor = maxY + 0.5;
    this.minZcor = minZ - 0.5;
    this.maxZcor = maxZ + 0.5;

    // The midpoints of the world, in world coords.
    this.centerX = (minX + maxX) / 2;
    this.centerY = (minY + maxY) / 2;
    this.centerZ = (minZ + maxZ) / 2;

    this.numPatches = this.numX * this.numY;
  }

  /**
   * Return the options (minX, maxX, ..) used to create this world.
   *
   * @returns {Object} A WorldOptions object
   */
  getOptions() {
    const { minX, minY, minZ, maxX, maxY, maxZ } = this;
    return { minX, minY, minZ, maxX, maxY, maxZ };
  }

  /**
   * Return a random 2D point within the World
   *
   * @returns {Array} A random x,y float array
   */
  randomPoint() {
    return [
      randomFloat2(this.minXcor, this.maxXcor),
      randomFloat2(this.minYcor, this.maxYcor),
    ];
  }

  /**
   * Return a random 3D point within the World
   *
   * @returns {Array} A random x,y,z float array
   */
  random3DPoint() {
    return [
      randomFloat2(this.minXcor, this.maxXcor),
      randomFloat2(this.minYcor, this.maxYcor),
      randomFloat2(this.minZcor, this.maxZcor),
    ];
  }

  /**
   * Return a random Patch 2D integer point
   *
   * @returns {Array}  A random x,y integer array
   */
  randomPatchPoint() {
    return [
      randomInt2(this.minX, this.maxX),
      randomInt2(this.minY, this.maxY),
    ];
  }

  /**
   * Given x,y,z values return true if within the world
   *
   * @param {number} x x value
   * @param {number} y y value
   * @param {number} [z=this.centerZ] z value
   * @returns {boolean} Whether or not on-world
   */
  isOnWorld(x, y, z = this.centerZ) {
    return (
      this.minXcor <= x &&
      x <= this.maxXcor &&
      //
      this.minYcor <= y &&
      y <= this.maxYcor &&
      //
      this.minZcor <= z &&
      z <= this.maxZcor
    );
  }
  // cropToWorld(x, y) {}

  // Note minX etc NOT the world's but of the coord sys we want to use.
  /**
   * Return an instance of a bounding box 2D transform.
   * It linearly interpolates between the given minX, minY, maxX, maxY,
   * and the world's values of the same properties.
   *
   * The parameters are in the popular geojson order: west, south, east, north
   *
   * Useful for Canvas top-left transforms and geojson transforms.
   *
   * @param {number} minX min bounding box x value
   * @param {number} minY min bounding box y value
   * @param {number} maxX max bounding box x value
   * @param {number} maxY max bounding box y value
   * @returns {BBoxTransform} Instance of the BBoxTransform
   */
  bboxTransform(minX, minY, maxX, maxY) {
    return new BBoxTransform(minX, minY, maxX, maxY, this);
  }

  // ### Following use PatchSize

  // Get the world size in pixels. PatchSize is optional, defalting to 1
  getWorldSize(patchSize = 1) {
    return [this.numX * patchSize, this.numY * patchSize];
  }

  // Convert a canvas context to world euclidean coordinates
  // Change the ctx.canvas size, determined by patchSize.
  setEuclideanTransform(ctx, patchSize) {
    // ctx.canvas.width = this.numX * patchSize
    // ctx.canvas.height = this.numY * patchSize
    this.setCanvasSize(ctx.canvas, patchSize);
    ctx.restore(); // close earlier save(). OK if no save called yet.
    ctx.save();
    ctx.scale(patchSize, -patchSize);
    ctx.translate(-this.minXcor, -this.maxYcor);
  }
  // Return patch size for given canvas.
  // Error if canvas patch width/height differ.
  patchSize(canvas) {
    const { numX, numY } = this;
    const { clientWidth: width, clientHeight: height } = canvas;
    const xSize = width / numX;
    const ySize = height / numY;
    if (xSize !== ySize) {
      throw Error(`World patchSize: x/y sizes differ ${xSize}, ${ySize}`);
    }
    return xSize;
  }
  // Change canvas size to this world's size.
  // Does not change size if already the same, preserving the ctx content.
  setCanvasSize(canvas, patchSize) {
    const [width, height] = this.getWorldSize(patchSize);
    setCanvasSize(canvas, width, height);
  }

  // Convert pixel location (top/left offset i.e. mouse) to patch coords (float)
  pixelXYtoPatchXY(x, y, patchSize) {
    return [this.minXcor + x / patchSize, this.maxYcor - y / patchSize];
  }
  // Convert patch coords (float) to pixel location (top/left offset i.e. mouse)
  patchXYtoPixelXY(x, y, patchSize) {
    return [(x - this.minXcor) * patchSize, (this.maxYcor - y) * patchSize];
  }

  xyToPatchIndex(x, y) {
    if (!this.isOnWorld(x, y)) return undefined;
    const { minX, maxX, maxY, numX, maxXcor, maxYcor } = this;
    x = x === maxXcor ? maxX : Math.round(x);
    y = y === maxYcor ? maxY : Math.round(y);
    return x - minX + numX * (maxY - y);
  }
  // patchIndexToXY(index) {}
}

/**
 * A linear transformer between world coords and the given bounding box.
 *
 * @class
 * @private
 */
class BBoxTransform {
  /**
   * Creates an instance of BBoxTransform.
   * See https://tools.ietf.org/html/rfc7946#section-5
   * .. which defines bbox as [west, south, east, north]
   * .. which is the bbox [minX, minY, maxX, maxY]
   * @param {number} minX min bounding box x value
   * @param {number} minY max bounding box x value
   * @param {number} maxX min bounding box y value
   * @param {number} maxY max bounding box y value
   * @param {World} world instance of a World object
   */
  constructor(minX, minY, maxX, maxY, world) {
    this.bbox = [minX, minY, maxX, maxY];

    if (minX < maxX) console.log("flipX");
    if (maxY < minY) console.log("flipY");

    if (minX < maxX) [minX, maxX] = [maxX, minX];
    if (maxY < minY) [maxY, minY] = [minY, maxY];
    const { maxXcor, maxYcor, minXcor, minYcor } = world;

    const mx = (minX - maxX) / (maxXcor - minXcor);
    const my = (maxY - minY) / (maxYcor - minYcor);

    const bx = (minX + maxX - mx * (maxXcor + minXcor)) / 2;
    const by = (maxY + minY - my * (maxYcor + minYcor)) / 2;

    Object.assign(this, { mx, my, bx, by });
  }

  /**
   * Convert from bbox point to world point
   *
   * @param {Array} bboxPoint A point in the bbox coordinates
   * @returns {Array} A point in the world coordinates
   */
  toWorld(bboxPoint) {
    const { mx, my, bx, by } = this;
    const [bboxX, bboxY] = bboxPoint;
    const x = (bboxX - bx) / mx;
    const y = (bboxY - by) / my;
    return [x, y];
  }

  /**
   * Convert from world point to bbox point
   *
   * @param {Array} worldPoint A point in the world coordinates
   * @returns {Array} A point in the bbox coordinates
   */
  toBBox(worldPoint) {
    const { mx, my, bx, by } = this;
    const [worldX, worldY] = worldPoint;
    const x = mx * worldX + bx;
    const y = my * worldY + by;
    return [x, y];
  }

  // aspectRatio() {
  //     const [west, south, east, north] = this.bbox
  //     return Math.abs((east - west) / (north - south))
  // }

  // Return center [x,y] of bbox in bbox coords.
  bboxCenter() {
    const [west, south, east, north] = this.bbox;
    return [(west + east) / 2, (south + north) / 2];
  }
  // Return 4 coordinate points of bbox corners, from topLeft, clockwise.
  bboxCoords() {
    const [west, south, east, north] = this.bbox;
    return [
      [west, north],
      [east, north],
      [east, south],
      [west, south],
    ];
  }
}

/**
 * A general color module, supporting css string colors, canvas2d pixel
 * colors, webgl and canvas2d Uint8ClampedArray r,g,b,a arrays.
 *
 * ### CSS Color Strings.
 *
 * CSS colors in HTML are strings, see [Mozillas Color Reference](
 * https://developer.mozilla.org/en-US/docs/Web/CSS/color_value),
 * taking one of 7 forms:
 *
 * * Names: over 140 color case-insensitive names like
 *   Red, Green, CadetBlue, etc.
 * * Hex, short and long form: #0f0, #ff10a0
 * * RGB: rgb(255, 0, 0), rgba(255, 0, 0, 0.5)
 * * HSL: hsl(120, 100%, 50%), hsla(120, 100%, 50%, 0.8)
 *
 * See [this wikipedia article](https://goo.gl/ev8Kw0)
 * on differences between HSL and HSB/HSV.
 */
const Color = {
  /**
   * Convert 4 r,g,b,a ints in [0-255] ("a" defaulted to 255) to a
   * css color string.
   */
  // Alpha "a" is converted to float in 0-1 for css string.
  // We use alpha in [0-255] to be compatible with TypedArray conventions.
  rgbaCssColor(r, g, b, a = 255) {
    a = a / 255;
    const a2 = a.toPrecision(2);
    return a === 1 ? `rgb(${r},${g},${b})` : `rgba(${r},${g},${b},${a2})`;
  },

  /**
   * Convert 4 ints, h,s,l,a, h in [0-360], s,l in [0-100]% a in [0-255] to a
   * css color string.
   *
   * NOTE: h=0 and h=360 are the same, use h in 0-359 for unique colors.
   */
  // Alpha "a" is converted to float in 0-1 for css string.
  hslCssColor(h, s, l, a = 255) {
    a = a / 255;
    const a4 = a.toPrecision(4);
    return a === 1 ? `hsl(${h},${s}%,${l}%)` : `hsla(${h},${s}%,${l}%,${a4})`;
  },

  // Return a html/css hex color string for an r,g,b opaque color (a=255).
  // Hex strings do not support alpha.
  //
  // Both #nnn and #nnnnnn forms supported.
  // Default is to check for the short hex form.
  hexCssColor(r, g, b) {
    return `#${(0x1000000 | (b | (g << 8) | (r << 16)))
      .toString(16)
      .slice(-6)}`;
  },

  // cssColor is a hybrid string and is our standard.  It returns:
  //
  // * rgbaCssColor if a not 255 (i.e. not opaque)
  // * hexCssColor otherwise
  cssColor(r, g, b, a = 255) {
    return a === 255
      ? this.hexCssColor(r, g, b, true)
      : this.rgbaCssColor(r, g, b, a);
  },

  randomCssColor() {
    const r255 = () => randomInt(256); // random int in [0,255]
    return this.cssColor(r255(), r255(), r255());
  },

  randomGrayCssColor(min = 0, max = 255) {
    const gray = randomInt2(min, max); // random int in [min,max]
    return this.cssColor(gray, gray, gray);
  },

  // ### Pixels

  cssToPixel(string) {
    const rgba = this.cssToUint8Array(string);
    return this.rgbaToPixel(...rgba);
  },

  rgbaToPixel(r, g, b, a = 255) {
    const rgba = new Uint8Array([r, g, b, a]);
    const pixels = new Uint32Array(rgba.buffer);
    return pixels[0];
  },

  randomPixel() {
    const r255 = () => randomInt(256); // random int in [0,255]
    return this.rgbaToPixel(r255(), r255(), r255());
  },

  randomGrayPixel(min = 0, max = 255) {
    const gray = randomInt2(min, max); // random int in [min,max]
    return this.rgbaToPixel(gray, gray, gray);
  },

  // ### CSS String Conversions

  // Return 4 element array given any legal CSS string color.
  //
  // Because strings vary widely: CadetBlue, #0f0, rgb(255,0,0),
  // hsl(120,100%,50%), we do not parse strings, instead we let
  // the browser do our work: we fill a 1x1 canvas with the css string color,
  // returning the r,g,b,a canvas ImageData TypedArray.

  // The shared 1x1 canvas 2D context.
  sharedCtx1x1: createCtx(1, 1), // share across calls.
  // Convert any css string to 4 element Uint8ClampedArray TypedArray.
  // If you need a JavaScript Array, use `new Array(...TypedArray)`
  // Slow, but works for all css strings: hsl, rgb, .. as well as names.
  cssToUint8Array(string) {
    this.sharedCtx1x1.clearRect(0, 0, 1, 1);
    this.sharedCtx1x1.fillStyle = string;
    this.sharedCtx1x1.fillRect(0, 0, 1, 1);
    return this.sharedCtx1x1.getImageData(0, 0, 1, 1).data;
  },

  // ### Typed Color
  // A TypedColor is a 4 element ArrayBuffer, with two views:
  //
  // * pixelArray: A single element Uint32Array view
  // * u8array: A 4 element r,g,b,a Uint8ClampedArray view
  //
  // getters/setters are provided for multiple other color types:
  //  'css', 'pixel', 'rgb', 'webgl'
  // If g is undefinec, returns toTypedColor(g)
  typedColor(r, g, b, a = 255) {
    if (g === undefined) return this.toTypedColor(r);
    const u8array = new Uint8ClampedArray([r, g, b, a]);
    u8array.pixelArray = new Uint32Array(u8array.buffer); // one element array
    // Make this an instance of TypedColorProto
    Object.setPrototypeOf(u8array, TypedColorProto);
    return u8array;
  },
  isTypedColor(any) {
    return any && any.constructor === Uint8ClampedArray && any.pixelArray;
  },

  // Return a typedColor given a value and optional colorType
  // If the value already is a typedColor, simply return it
  // If colorType not defined, assume css (string) or pixel (number)
  // The colorType can be: 'css', 'pixel', 'rgb', 'webgl'
  // Note rgb and webgl are int arrays & float arrays respectively.
  toTypedColor(value, colorType) {
    if (this.isTypedColor(value)) return value;

    const tc = this.typedColor(0, 0, 0, 0); // "empty" typed color
    if (colorType == null) {
      if (typeof value === "string") tc.css = value;
      else if (typeof value === "number") tc.pixel = value;
      else throw Error(`toTypedColor: illegal value ${value}`);
    } else {
      // REMIND: type check value & colorType?
      tc[colorType] = value;
    }
    return tc;
  },

  // Random typedColor, rgb or gray, alpha = 255 for both:
  randomTypedColor() {
    const r255 = () => randomInt(256); // random int in [0,255]
    return this.typedColor(r255(), r255(), r255());
  },
  // Random gray color, alpha = 255
  randomGrayTypedColor(min = 0, max = 255) {
    const gray = randomInt2(min, max); // random int in [min,max]
    return this.typedColor(gray, gray, gray);
  },
  // Arrays of random typed colors, very useful for patch colors
  randomColorArray(length) {
    const colors = new Array(length);
    forLoop(colors, (c, i) => (colors[i] = this.randomTypedColor()));
    return colors;
  },
  randomGrayArray(length, min = 0, max = 255) {
    const grays = new Array(length);
    forLoop(
      grays,
      (g, i) => (grays[i] = this.randomGrayTypedColor(min, max))
    );
    return grays;
  },
};

// Prototype for Color. Getters/setters for usability, may be slower.
const TypedColorProto = {
  // Inherit from Uint8ClampedArray
  __proto__: Uint8ClampedArray.prototype,

  // Set the Color to new rgba values.
  setColor(r, g, b, a = 255) {
    this.checkColorChange();
    this[0] = r;
    this[1] = g;
    this[2] = b;
    this[3] = a;
  },
  // No real need for getColor, it *is* the typed Uint8 array
  set rgb(rgbaArray) {
    this.setColor(...rgbaArray);
  },
  get rgb() {
    return this;
  },

  // Set opacity to a value in 0-255
  setAlpha(alpha) {
    this.checkColorChange();
    this[3] = alpha; // Uint8ClampedArray will clamp to 0-255
  },
  getAlpha() {
    return this[3];
  },
  get alpha() {
    return this.getAlpha();
  },
  set alpha(alpha) {
    this.setAlpha(alpha);
  },

  // Set the Color to a new pixel value
  setPixel(pixel) {
    this.checkColorChange();
    this.pixelArray[0] = pixel;
  },
  // Get the pixel value
  getPixel() {
    return this.pixelArray[0];
  },
  get pixel() {
    return this.getPixel();
  },
  set pixel(pixel) {
    this.setPixel(pixel);
  },

  // Set pixel/rgba values to equivalent of the css string.
  // 'red', '#f00', 'ff0000', 'rgb(255,0,0)', etc.
  //
  // Does *not* set the chached this.string, which will be lazily evaluated
  // to its common cssColor by getCss(). The above would all return '#f00'.
  setCss(string) {
    return this.setColor(...Color.cssToUint8Array(string));
  },
  // Return the cssColor for this Color, cached in the @string value
  getCss() {
    if (this.string == null) this.string = Color.cssColor(...this);
    return this.string;
  },
  get css() {
    return this.getCss();
  },
  set css(string) {
    this.setCss(string);
  },

  // Note: webgl colors are 3 RGB floats (no A)
  setWebgl(array) {
    if (array.length !== 3)
      throw Error("setWebgl array length must be 3, length:", array.length);
    this.setColor(
      // OK if float * 255 non-int, setColor stores into uint8 array
      array[0] * 255,
      array[1] * 255,
      array[2] * 255
    );
  },
  getWebgl() {
    return [this[0] / 255, this[1] / 255, this[2] / 255];
  },
  get webgl() {
    return this.getWebgl();
  },
  set webgl(array) {
    this.setWebgl(array);
  },

  // Housekeeping when the color is modified.
  checkColorChange() {
    // Reset string & webgl on color change.
    this.string = null; // will be lazy evaluated via getCss.
    this.floatArray = null;
  },
  // Return true if color is same value as myself, comparing pixels
  equals(color) {
    return this.getPixel() === color.getPixel();
  },
  toString() {
    return `[${Array.from(this).toString()}]`;
  },
  // Return a [distance metric](
  // http://www.compuphase.com/cmetric.htm) between two colors.
  // Max distance is roughly 765 (3*255), for black & white.
  // For our purposes, omitting the sqrt will not effect our results
  rgbDistance(r, g, b) {
    const [r1, g1, b1] = this;
    const rMean = Math.round((r1 + r) / 2);
    const [dr, dg, db] = [r1 - r, g1 - g, b1 - b];
    const [dr2, dg2, db2] = [dr * dr, dg * dg, db * db];
    const distanceSq =
      (((512 + rMean) * dr2) >> 8) + 4 * dg2 + (((767 - rMean) * db2) >> 8);
    return distanceSq; // Math.sqrt(distanceSq)
  },
};

// const getPixel = color => color.pixel || color
function getPixel(color) {
  if (typeof pixel === "number") return pixel;
  if (color.pixel) return color.pixel;
  return Color.toTypedColor(color).pixel;
}

class PatchesView {
  // Ctor: create a 2D context and imageData for this View
  constructor(width, height) {
    this.ctx = createCtx(width, height);
    this.resetImageData();
    // this.imageData = util.ctxImageData(this.ctx)
    // this.pixels = new Uint32Array(this.imageData.data.buffer)
    // this.length = this.pixels.length
    this.useImageSmoothing = false;
  }
  // Set the imageData and pixels values from the pixel's canvas ctx
  resetImageData() {
    this.imageData = ctxImageData(this.ctx);
    this.pixels = new Uint32Array(this.imageData.data.buffer);
    // this.length = this.pixels.length
  }
  setPatchesSmoothing(smoothting) {
    this.useImageSmoothing = smoothting;
  }

  // Install pixels in this imageData object.
  // pixelFcn(d) returns a pixel for each data item.data can be patches
  // or data derived from patches using patch state values.
  setPixels(data, pixelFcn = (d) => d) {
    if (isOofA(data)) data = toAofO(data);
    if (data.length !== this.pixels.length) {
      throw Error(
        "setPixels, data.length != pixels.length " +
          data.length +
          " " +
          this.pixels.length
      );
    }
    forLoop(data, (d, i) => {
      this.pixels[i] = getPixel(pixelFcn(d));
    });

    // if (updateCanvas) this.ctx.putImageData(this.imageData, 0, 0)
  }
  createPixels(pixelFcn) {
    repeat(this.pixels.length, (i) => {
      this.pixels[i] = getPixel(pixelFcn(i));
    });
    // if (updateCanvas) this.ctx.putImageData(this.imageData, 0, 0)
  }
  // Used to be: setPixel(x, y, pixel) {
  // but best to be purely independent of world object
  setPixel(index, pixel) {
    // const index = world.xyToPatchIndex(x, y)
    this.pixels[index] = getPixel(pixel);
  }

  // Draw this pixel canvas onto a View 2D canvas ctx.
  draw(ctx) {
    const smoothing = this.ctx.imageSmoothingEnabled;
    ctx.imageSmoothingEnabled = this.useImageSmoothing;
    this.ctx.putImageData(this.imageData, 0, 0);
    // this.updateCanvas()
    fillCtxWithImage(ctx, this.ctx.canvas);
    ctx.imageSmoothingEnabled = smoothing;
  }
  clear(color) {
    // typedColor -> css
    color = color.css || color;

    if (!color || typeof color === "string") {
      clearCtx(this.ctx, color);
    } else if (typeof color === "number") {
      this.createPixels(() => color);
    } else {
      throw Error("patchesView.clear(): illegal color " + color);
    }

    if (typeof color === "number") {
      this.updateCanvas();
    } else {
      this.resetImageData();
    }
  }

  // Return promise for an ImageBitmap of the current ctx
  // Note safari does not support createImageBitmap as of 4/20/19
  // REMIND: See if imagebitmaps can avoid img alpha premultiply etc
  getImageBitmap(options = {}) {
    return createImageBitmap(this.imageData, options);
  }
  drawImageBitmap(ctx, options = {}) {
    createImageBitmap(this.imageData, options).then((img) =>
      fillCtxWithImage(ctx, img)
    );
  }

  // Push imageData to canvas, return the canvas
  updateCanvas() {
    this.ctx.putImageData(this.imageData, 0, 0);
    return this.ctx.canvas;
  }
}

function cssColor$1(color) {
  if (color) return color.css || color;
  return color;
}

// Shapes are Canvas2D drawings in -0.5 to +0.5, unit squares
// They are drawn on a Canvas2D using transforms for x, y, theta
class Shapes {
  constructor() {
    this.cache = {};
    this.paths = paths; // For accessing the paths thru the Shapes instance
  }
  // Add a new path. Error if name already used
  addPath(name, pathFunction) {
    if (this.getPath(name)) ;
    paths[name] = pathFunction;
  }
  // NOTE: convention: if shapeName ends in 2, then needs strokeColor
  needsStrokeColor(shapeName) {
    return shapeName.endsWith("2");
  }

  // Get an array of path names. Can change over time via addPath()
  getPathNames() {
    return Object.keys(paths);
  }
  // Get a path drawing function by name. Returns 'undefined' if not found.
  getPath(name) {
    return paths[name];
  }

  // Return random shape function. It's name is shape.name
  oneOf() {
    // return paths[util.oneValOf(this.getPathNames())]
    return oneValOf(paths);
  }

  // Return the name/path at index of path array.
  // Wrap the index to be within the array.
  nameAtIndex(index) {
    const names = this.getPathNames();
    return names[mod(index, names.length)];
  }
  atIndex(index) {
    const name = this.nameAtIndex(index);
    return paths[name];
  }

  imagePathPromise(name, imgURL) {
    return imagePromise(imgURL).then((img) => {
      this.createImagePath(name, img);
    });
  }

  // Create a shape that is an imageable (img, canvas, ...)
  createImagePath(name, img, flip = true) {
    if (!isImageable(img)) {
      throw Error("Shapes createImagePath: img not an imageable " + img);
    }
    if (flip) img = flipImage(img);
    function imagePath(ctx) {
      ctx.drawImage(img, -0.5, -0.5, 1, 1);
    }
    // paths[name] = imagePath
    this.addPath(name, imagePath);
  }

  imageName(name, pixels, fill, stroke) {
    const path = this.getPath(name);
    if (!Number.isInteger(pixels))
      throw Error(`imageName: pixels is not integer: ${name}`);
    if (!path) throw Error(`imageName: ${name} not in Shapes`);

    if (path.name === "imagePath") return `${name}_${pixels}_image`;

    if (!fill) throw Error(`imageName: No color for shape ${name}`);
    // OK to give strokeColor when not needed.
    if (!this.needsStrokeColor(name)) stroke = null;

    return `${name}_${pixels}_${fill}${stroke ? `_${stroke}` : ""}`;
  }

  shapeToImage(name, pixels, fill, stroke) {
    pixels = Math.ceil(pixels);
    const imgName = this.imageName(name, pixels, fill, stroke);
    if (this.cache && this.cache[imgName]) return this.cache[imgName];

    const ctx = createCtx(pixels, pixels);
    ctx.fillStyle = cssColor$1(fill);
    ctx.strokeStyle = cssColor$1(stroke);
    ctx.scale(pixels, -pixels);
    ctx.translate(0.5, -0.5);
    ctx.beginPath();
    paths[name](ctx);
    ctx.closePath();
    ctx.fill();

    ctx.canvas.name = imgName;
    if (this.cache) this.cache[imgName] = ctx.canvas;
    return ctx.canvas;
  }

  imageNameToImage(imageName) {
    const [name, pixels, fill, stroke] = imageName.split("_");
    return this.shapeToImage(name, pixels, fill, stroke);
  }
}

// Utilities:
// Flip an image so will be upright in euclidean coords
function flipImage(img) {
  const { width, height } = img;
  const ctx = createCtx(width, height);
  ctx.scale(1, -1);
  ctx.drawImage(img, 0, -height);
  return ctx.canvas;
}
// Centered polygon, x,y in [-1, 1] (size = 2)
function poly(ctx, points) {
  points.forEach((pt, i) => {
    if (i === 0) ctx.moveTo(pt[0], pt[1]);
    else ctx.lineTo(pt[0], pt[1]);
  });
}
// centered circle
function circle(ctx, x, y, radius, anticlockwise = false) {
  ctx.arc(x, y, radius, 0, 2 * Math.PI, anticlockwise);
}
// centered square
function square(ctx, x, y, size) {
  ctx.fillRect(x - size / 2, y - size / 2, size, size);
}

// The paths object containing shape path procedures for common shapes.
// Paths are unit size, -.5 -> +.5
// Use "addPath()" to add your own new shapes.
//
// These are "upsidedown", i.e. not upright. This is so we can traansform
// to euclidean coords.
// Does not impact most paths. But will impact images & person, person2
const paths = {
  // default(ctx) {
  //     this.dart(ctx)
  // },
  arrow(ctx) {
    poly(ctx, [
      [0.5, 0],
      [0, 0.5],
      [0, 0.2],
      [-0.5, 0.2],
      [-0.5, -0.2],
      [0, -0.2],
      [0, -0.5],
    ]);
  },
  bug(ctx) {
    ctx.strokeStyle = ctx.fillStyle;
    this.bug2(ctx);
  },
  bug2(ctx) {
    ctx.lineWidth = 0.05;
    poly(ctx, [
      [0.4, 0.225],
      [0.2, 0],
      [0.4, -0.225],
    ]);
    ctx.stroke();
    ctx.beginPath();
    circle(ctx, 0.12, 0, 0.13);
    circle(ctx, -0.05, 0, 0.13);
    circle(ctx, -0.27, 0, 0.2);
  },
  circle(ctx) {
    // ctx.arc(0, 0, 1, 0, 2 * Math.PI)
    circle(ctx, 0, 0, 0.5);
  },
  dart(ctx) {
    poly(ctx, [
      [0.5, 0],
      [-0.5, 0.4],
      [-0.25, 0],
      [-0.5, -0.4],
    ]);
  },
  frame(ctx) {
    const inset = 0.2;
    const r = 0.5 - inset;
    poly(ctx, [
      [-0.5, -0.5],
      [0.5, -0.5],
      [0.5, 0.5],
      [-0.5, 0.5],
    ]); //cclockwise
    ctx.closePath();
    // reverse direction for non-zero winding rule to leave empty center rect.
    poly(ctx, [
      [-r, -r],
      [-r, r],
      [r, r],
      [r, -r],
    ]); // clockwise
  },
  frame2(ctx) {
    const inset = 0.2;
    square(ctx, 0, 0, 1);
    ctx.fillStyle = ctx.strokeStyle;
    square(ctx, 0, 0, 1 - 2 * inset); //2 - 2 * inset)
  },
  person(ctx) {
    ctx.strokeStyle = ctx.fillStyle;
    this.person2(ctx);
  },
  person2(ctx) {
    poly(ctx, [
      [0.15, 0.2],
      [0.3, 0],
      [0.125, -0.1],
      [0.125, 0.05],
      [0.1, -0.15],
      [0.25, -0.5],
      [0.05, -0.5],
      [0, -0.25],
      [-0.05, -0.5],
      [-0.25, -0.5],
      [-0.1, -0.15],
      [-0.125, 0.05],
      [-0.125, -0.1],
      [-0.3, 0],
      [-0.15, 0.2],
    ]);
    ctx.closePath();
    ctx.fill();
    ctx.beginPath();
    ctx.fillStyle = ctx.strokeStyle;
    circle(ctx, 0, 0.35, 0.15);
  },
  ring(ctx) {
    const [rOuter, rInner] = [0.5, 0.3];
    circle(ctx, 0, 0, rOuter);
    ctx.lineTo(rInner, 0);
    circle(ctx, 0, 0, rInner, true);
  },
  ring2(ctx) {
    // fileStyle is outer color, strokeStyle inner color
    const [rOuter, rInner] = [0.5, 0.3];
    circle(ctx, 0, 0, rOuter);
    ctx.closePath();
    ctx.fill();
    ctx.beginPath();
    ctx.fillStyle = ctx.strokeStyle;
    circle(ctx, 0, 0, rInner);
  },
  square(ctx) {
    square(ctx, 0, 0, 1);
  },
  triangle(ctx) {
    poly(ctx, [
      [0.5, 0],
      [-0.5, -0.4],
      [-0.5, 0.4],
    ]);
  },
};

// export default Shapes

/*
ToDo:
    - simplify async images: make a "slot" then fill it.
    - manage inversion/upright. Only needed for: person/2, images
    - optimize:
        - cache images .. like spritesheet
        - use shortcuts via the helper functions (avoid transforms)
          See shapes.coffee in as0
    - paths: use object w/ meta-data, & draw(). needsStroke, for example
*/

function cssColor(color) {
  if (color) return color.css || color;
  return color;
}

class TurtlesView {
  static defaultOptions() {
    return {
      // canvasStack: false,
      useSprites: false,
      patchSize: 10,
    };
  }

  // constructor(ctx, patchSize, world, useSprites = false) {
  constructor(ctx, world, options = {}) {
    options = Object.assign(TurtlesView.defaultOptions(), options);
    Object.assign(this, { ctx, world }, options);
    this.shapes = new Shapes();
    this.reset(this.patchSize, this.useSprites);
  }
  reset(patchSize, useSprites = this.useSprites) {
    this.useSprites = useSprites;
    this.resetCtx(patchSize); // sets this.patchSize
  }

  getImageBitmap() {
    return createImageBitmap(this.ctx.canvas);
  }

  resetCtx(patchSize) {
    this.patchSize = patchSize;
    if (this.useSprites) {
      // If using sprites, do not install the transform
      this.world.setCanvasSize(this.ctx.canvas, patchSize);
      // Fix for sticky euclidean xfm if no size change
      this.ctx.restore();
      // util.setIdentity(this.ctx)
    } else {
      // If using shapes, create euclidean transform.
      this.world.setEuclideanTransform(this.ctx, patchSize);
    }
  }

  drawTurtles(data, viewFcn) {
    if (isOofA(data)) data = toAofO(data);
    const constantView = isObject(viewFcn);
    forLoop(data, (turtle, i) => {
      const viewData = constantView ? viewFcn : viewFcn(turtle, i, data);
      this.drawTurtle(turtle, viewData);
    });
  }
  drawTurtle(turtle, viewData) {
    if (turtle.hidden) return;
    if (viewData.size === 0) return; // skip this turtle if size === 0
    if (this.useSprites) {
      let { sprite, noRotate } = viewData;
      // If not the short form, create the image.
      // Not too bad, it uses the shapes.cache after first image.
      if (!sprite) {
        const { shape, color, strokeColor, size } = viewData;
        const pixels = size * this.patchSize;
        sprite = this.shapes.shapeToImage(shape, pixels, color, strokeColor);
      }

      if (sprite.sheet) {
        sprite.sheet.draw(
          this.ctx,
          sprite,
          turtle.x,
          turtle.y,
          turtle.theta,
          this.world,
          this.patchSize,
          noRotate
        );
      } else {
        this.drawImage(sprite, turtle.x, turtle.y, noRotate ? 0 : turtle.theta);
      }
    } else {
      const { shape, color, strokeColor, size, noRotate } = viewData;
      this.drawShape(
        shape,
        turtle.x,
        turtle.y,
        noRotate ? 0 : turtle.theta,
        size,
        color,
        strokeColor
      );
    }
  }

  drawShape(name, x, y, theta = 0, size = 1, fill, stroke) {
    const ctx = this.ctx;
    ctx.save();

    // note: neither may be needed! image needs none, no-stroke only fill.
    ctx.fillStyle = cssColor(fill); // fill.css || fill
    ctx.strokeStyle = cssColor(stroke); // stroke.css || stroke

    ctx.translate(x, y);
    ctx.scale(size, size);
    if (theta !== 0) ctx.rotate(theta);
    ctx.beginPath();
    this.shapes.paths[name](ctx);
    ctx.closePath();
    ctx.fill();

    ctx.restore();
  }

  drawImage(img, x, y, theta = 0) {
    const halfPix = img.width / 2;
    const patchSize = this.patchSize; //this.world.canvasPatchSize(ctx.canvas)
    const [x0, y0] = this.world.patchXYtoPixelXY(x, y, patchSize);
    const ctx = this.ctx;

    if (theta === 0) {
      ctx.drawImage(img, x0 - halfPix, y0 - halfPix);
    } else {
      ctx.save();
      ctx.translate(x0, y0);
      ctx.rotate(-theta);
      ctx.drawImage(img, -halfPix, -halfPix);
      ctx.restore();
    }
  }

  drawLinks(data, viewFcn) {
    if (isOofA(data)) data = toAofO(data);
    const uniformLinks = isObject(viewFcn);
    const ctx = this.ctx;
    setIdentity(this.ctx);
    if (uniformLinks) {
      ctx.strokeStyle = cssColor(viewFcn.color);
      ctx.lineWidth = viewFcn.width || 1;
      ctx.beginPath();
    }
    forLoop(data, (link, i) => {
      if (uniformLinks) {
        this.drawLink(link);
      } else {
        const { color, width } = viewFcn(link, i, data);
        this.drawLink(link, color, width);
      }
    });
    if (uniformLinks) {
      ctx.closePath();
      ctx.stroke();
    }
    this.ctx.restore(); // set identity does a save()
  }

  drawLink(link, color, width = 1) {
    this.drawLine(link.x0, link.y0, link.x1, link.y1, color, width);
  }

  drawLine(x0, y0, x1, y1, stroke, width = 1) {
    // ctx.save() // set identity does a save()
    const ctx = this.ctx;
    [x0, y0] = this.world.patchXYtoPixelXY(x0, y0, this.patchSize);
    [x1, y1] = this.world.patchXYtoPixelXY(x1, y1, this.patchSize);

    if (stroke) {
      ctx.strokeStyle = cssColor(stroke);
      ctx.lineWidth = width;
      ctx.beginPath();
    }
    ctx.moveTo(x0, y0);
    ctx.lineTo(x1, y1);
    // util.setIdentity(ctx)
    if (stroke) {
      ctx.closePath();
      ctx.stroke();
    }
    // ctx.restore()
  }
}

/*

ToDo:
    - hello: *faster* w/o sprites! (uniform links)
        http://localhost/src/agentscript/tests/hello.html 144
        http://localhost/src/agentscript/tests/hello.html?sprites 95
      Possibly the sprite name lookup is slow? w/ non-uniform links:
        http://localhost/src/agentscript/tests/hello.html?links 70
        http://localhost/src/agentscript/tests/hello.html?links&sprites 83
      !! now faster w/ sprites.
      Gotta be a bug of some sort
    - hello: noTurtles: uniform links: 1068, non: 205! fast!
    - hello: noLinks: 149, sprites: 118! What's up? Darts faster than imgs?

    - ctor: change params,
    - The OofA needs parameters/template, not just x,y,theta

*/

class TwoView {
  static defaultOptions() {
    return {
      div: document.body,
      useSprites: false,
      patchSize: 10,
      // width: null,
    };
  }

  // ======================

  // Note: world can be a model, who'd model.world will be used
  constructor(world, options = {}) {
    // options: override defaults, assign to this
    // Object.assign(this, TwoView.defaultOptions(), options)
    options = Object.assign(TwoView.defaultOptions(), options);
    if (options.width) {
      options.patchSize = options.width / world.width;
      delete options.width;
    }
    this.div = options.div;

    // let div = this.div
    let div = options.div;
    let can = div;
    div = isString(div) ? document.getElementById(div) : div;
    if (!isCanvas(can)) {
      can = createCanvas(0, 0, false); // not offscreen
      div.appendChild(can);
      // div = can
    }
    // this.div = div

    this.ctx = can.getContext("2d");
    this.world = new World(world.world || world); // world can be model

    // Object.assign(this, { ctx, world }, options)
    // Object.assign(this, { ctx, world })

    this.patchesView = new PatchesView(this.world.width, this.world.height);
    this.turtlesView = new TurtlesView(this.ctx, this.world, options);

    this.ticks = 0;
    this.clear();
  }
  tick() {
    this.ticks++;
  }
  get canvas() {
    return this.ctx.canvas;
  }
  // setPatchSize(patchSize) {
  //     this.reset(patchSize)
  // }
  // setWidth(width) {
  //     this.reset(width / this.world.width)
  // }
  // get patchSize() {
  //     this.turtlesView.patchSize
  // }
  reset(patchSize, useSprites = this.useSprites) {
    // if (this.width) this.width = patchSizeOrWidth
    // this.patchSize = this.width
    //     ? this.width / this.world.width
    //     : patchSizeOrWidth
    // this.patchSize = patchSize
    // this.useSprites = useSprites
    this.turtlesView.reset(patchSize, useSprites);
  }

  get width() {
    // return this.world.width * this.turtlesView.patchSize // this.patchSize
    return this.world.width * this.patchSize;
  }
  set width(val) {
    this.reset(val / this.world.width);
  }
  get patchSize() {
    return this.turtlesView.patchSize;
  }
  set patchSize(val) {
    this.reset(val);
  }
  get useSprites() {
    return this.turtlesView.useSprites;
  }
  set useSprites(val) {
    this.reset(this.patchSize, val);
  }

  // Clear the view.canvas.
  // If no color, or 'transparent', make transparent via clearCtx.
  // If cssColor.css (typedColor) use it, else cssColor
  clear(cssColor) {
    clearCtx(this.ctx, cssColor);

    // REMIND: Clear the pixels?
    // this.patchesView.clear(cssColor)
  }

  createPatchPixels(pixelFcn) {
    this.patchesView.createPixels(pixelFcn);
  }
  // setPatchPixel(x, y, pixel) {
  //     this.patchesView.setPixel(x, y, pixel)
  // }
  setPatchPixel(index, pixel) {
    this.patchesView.setPixel(index, pixel);
  }
  setPatchesPixels(data, pixelFcn) {
    this.patchesView.setPixels(data, pixelFcn);
  }
  setPatchesSmoothing(smoothing) {
    this.patchesView.setPatchesSmoothing(smoothing);
  }
  drawPatchesImage(img) {
    // if (img) util.fillCtxWithImage(this.ctx, img) ??
    fillCtxWithImage(this.ctx, img);
  }

  // If no data, redraw with existing patchesView cache
  drawPatches(data, pixelFcn) {
    if (data != null) {
      this.patchesView.setPixels(data, pixelFcn);
    }
    this.patchesView.draw(this.ctx);
  }

  drawTurtles(data, viewFcn) {
    this.turtlesView.drawTurtles(data, viewFcn);
  }

  drawLinks(data, viewFcn) {
    this.turtlesView.drawLinks(data, viewFcn);
  }

  setTextProperties(font, textAlign = "center", textBaseline = "middle") {
    if (typeof font === "number")
      font = `${this.patchSize * font}px sans-serif`;
    setTextProperties(this.ctx, font, textAlign, textBaseline);
  }
  drawText(string, x, y, color = "black") {
    [x, y] = this.world.patchXYtoPixelXY(x, y, this.patchSize);
    string = "" + string; // convert numeric values to strings
    drawText(this.ctx, string, x, y, color);
  }
}

// A colormap is simply an array of typedColors with several utilities such

const ColorMap = {
  // ### Color Array Utilities
  // Several utilities for creating color arrays

  // ### Gradients

  // Ask the browser to use the canvas gradient feature
  // to create nColors given the gradient color stops and locs.
  // See Mozilla [Gradient Doc](
  // https://developer.mozilla.org/en-US/docs/Web/CSS/linear-gradient),
  //
  // This is a powerful browser feature, can be
  // used to create all the MatLab colormaps.
  //
  // Stops are css strings.
  // Locs are floats from 0-1, default is equally spaced.
  gradientImageData(nColors, stops, locs) {
    // Convert the color stops to css strings
    // stops = stops.map(c => cssColor(c))
    const ctx = createCtx(nColors, 1);
    // Install default locs if none provide
    if (!locs) locs = floatRamp(0, 1, stops.length);
    // Create a new gradient and fill it with the color stops
    const grad = ctx.createLinearGradient(0, 0, nColors, 0);
    repeat(stops.length, (i) => grad.addColorStop(locs[i], stops[i]));
    // Draw the gradient, returning the image data TypedArray
    ctx.fillStyle = grad;
    ctx.fillRect(0, 0, nColors, 1);
    return ctxImageData(ctx).data;
  },

  // ### Array Conversion Utilities

  // Convert a Uint8Array into Array of 4 element typedColors.
  // Useful for converting ImageData objects like gradients to colormaps.
  // WebGL ready: the array.typedArray is suitable for Uniforms.
  typedArraytoColors(typedArray) {
    const array = [];
    step(
      typedArray.length,
      4,
      // Note: can't share subarray as color's typed array:
      // it's buffer is for entire array, not just subarray.
      (i) => array.push(Color.typedColor(...typedArray.subarray(i, i + 4)))
    );
    array.typedArray = typedArray;
    return array;
  },
  // Convert an Array of Arrays to an Array of typedColors.
  // Webgl ready as above.
  arraysToColors(array) {
    const typedArray = new Uint8ClampedArray(array.length * 4);
    repeat(array.length, (i) => {
      const a = array[i];
      if (a.length === 3) a.push(255);
      typedArray.set(a, i * 4);
    });
    return this.typedArraytoColors(typedArray);
  },

  // Permute the values of 3 arrays. Ex:
  //
  // [1,2],[3],[4,5] -> [ [1,3,4],[1,3,5],[2,3,4],[2,3,5] ]
  permuteArrays(A1, A2 = A1, A3 = A1) {
    const array = [];
    for (const a3 of A3) {
      // sorta odd const works with ths, but...
      for (const a2 of A2) for (const a1 of A1) array.push([a1, a2, a3]);
    }
    return array;
  },
  // Use permuteArrays to create uniformly spaced color ramp permutation.
  // Ex: if numRs is 3, permuteArrays's A1 would be [0, 127, 255]
  permuteRGBColors(numRs, numGs = numRs, numBs = numRs) {
    const toRamp = (num) => integerRamp(0, 255, num);
    const ramps = [numRs, numGs, numBs].map(toRamp);
    return this.permuteArrays(...ramps);
  },

  // ### ColorMaps

  // ColorMaps are Arrays of TypedColors with these additional methods. Webgl
  // ready if made w/ `typedArraytoColors` or `arraysToColors` above.
  // Used to be memory effecent (shared colors), webgl compatible,  and for
  // MatLab-like color-as-data.
  ColorMapProto: {
    // Inherit from Array
    __proto__: Array.prototype,
    // Create a [sparse array](https://goo.gl/lQlq5k) of index[pixel] = pixel.
    // Used by indexOf below for exact match of a color within the colormap.
    createIndex() {
      this.index = [];
      repeat(this.length, (i) => {
        const px = this[i].getPixel();
        this.index[px] = i;
        if (this.cssNames) this.index[this.cssNames[i]] = i;
      });
    },
    // Return a random color within the colormap
    randomColor() {
      return this[randomInt(this.length)];
    },

    // Set alpha of all the colors ih the map
    // Note this will be shared by all users of this map!
    // Use clone() to have your own copy of a shared map.
    setAlpha(alpha) {
      forLoop(this, (color) => color.setAlpha(alpha));
    },

    // Clone this colorMap
    clone() {
      return ColorMap.cloneColorMap(this);
    },

    // Return the color at index of this array.
    // Wrap the index to be within the array.
    atIndex(index) {
      return this[index % this.length];
    },
    // Return the index of a typedColor within the colormap,
    // undefined if no exact match.
    // Use the `closest` methods below for nearest, not exact, match.
    indexOf(color) {
      if (this.index) return this.index[color.getPixel()];
      for (let i = 0; i < this.length; i++) {
        if (color.equals(this[i])) return i;
      }
      return undefined;
    },
    // Return color scaled by number within [min, max].
    // A linear interpolation (util.lerp) in [0, length-1].
    // Used to match data directly to a color as in MatLab.
    //
    // Ex: scaleColor(25, 0, 50) returns the color in the middle of the colormap
    scaleColor(number, min, max) {
      // number = util.clamp(number, min, max)
      const scale = lerpScale(number, min, max);
      const index = Math.round(lerp(0, this.length - 1, scale));
      return this[index];
    },
    // Return the Uint8 array used to create the typedColors,
    // undefined if not webgl ready.
    // webglArray() {
    //     return this.typedArray
    // },

    // Debugging: Return a string with length and array of colors
    toString() {
      return `${this.length} ${arraysToString(this)}`;
    },

    // Iterate through the colormap colors, returning the index of the
    // min typedColor.rgbDistance value from r, g, b
    rgbClosestIndex(r, g, b) {
      let minDist = Infinity;
      let ixMin = 0;
      for (var i = 0; i < this.length; i++) {
        const d = this[i].rgbDistance(r, g, b);
        if (d < minDist) {
          minDist = d;
          ixMin = i;
          if (d === 0) return ixMin;
        }
      }
      return ixMin;
    },
    // Return the color with the rgbClosestIndex value
    rgbClosestColor(r, g, b) {
      return this[this.rgbClosestIndex(r, g, b)];
    },

    // Calculate the closest cube index for the given r, g, b values.
    // Faster than rgbClosestIndex, does direct calculation, not iteration.
    cubeClosestIndex(r, g, b) {
      const cube = this.cube;
      const rgbSteps = cube.map((c) => 255 / (c - 1));
      const rgbLocs = [r, g, b].map((c, i) => Math.round(c / rgbSteps[i]));
      const [rLoc, gLoc, bLoc] = rgbLocs;
      return rLoc + gLoc * cube[0] + bLoc * cube[0] * cube[1];
    },
    cubeClosestColor(r, g, b) {
      return this[this.cubeClosestIndex(r, g, b)];
    },

    // Choose the appropriate method for finding closest index.
    // Lets the user specify any color, and let the colormap
    // use the best match.
    closestIndex(r, g, b) {
      return this.cube
        ? this.cubeClosestIndex(r, g, b)
        : this.rgbClosestIndex(r, g, b);
    },
    // Choose the appropriate method for finding closest color
    closestColor(r, g, b) {
      return this[this.closestIndex(r, g, b)];
    },
  },

  // ### Utilities for constructing ColorMaps

  // Convert an array of rgb(a) Arrays or TypedColors to a webgl-ready colormap.
  basicColorMap(colors) {
    colors = this.arraysToColors(colors);
    Object.setPrototypeOf(colors, this.ColorMapProto);
    return colors;
  },
  // Create a gray map (gray: r=g=b)
  // These are typically 256 entries but can be smaller
  // by passing a size parameter and the min/max range.
  grayColorMap(min = 0, max = 255, size = max - min + 1) {
    const ramp = integerRamp(min, max, size);
    return this.basicColorMap(ramp.map((i) => [i, i, i]));
  },

  // Create a colormap by permuted rgb values.
  //
  // numRs, numGs, numBs are numbers, the number of steps beteen 0-255.
  // Ex: numRs = 3, corresponds to 0, 128, 255.
  // NOTE: the defaults: rgbColorCube(6) creates a `6 * 6 * 6` cube.
  rgbColorCube(numRs, numGs = numRs, numBs = numRs) {
    const array = this.permuteRGBColors(numRs, numGs, numBs);
    const map = this.basicColorMap(array);
    // Save the parameters for fast color calculations.
    map.cube = [numRs, numGs, numBs];
    return map;
  },
  // Create a colormap by permuting the values of the given arrays.
  // Similar to above but with arrays that may have arbitrary values.
  rgbColorMap(R, G, B) {
    const array = this.permuteArrays(R, G, B);
    return this.basicColorMap(array);
  },

  // Create an hsl map, inputs are arrays to be permutted like rgbColorMap.
  // Convert the HSL values to Color.colors, default to bright hue ramp (L=50).
  hslColorMap(H, S = [100], L = [50]) {
    const hslArray = this.permuteArrays(H, S, L);
    const array = hslArray.map((a) =>
      Color.toTypedColor(Color.hslCssColor(...a))
    );
    return this.basicColorMap(array);
  },

  // Use gradient to build an rgba array, then convert to colormap.
  // Stops are css strings or typedColors.
  // locs defaults to evenly spaced, probably what you want.
  //
  // This easily creates all the MatLab colormaps like "jet" below.
  gradientColorMap(nColors, stops, locs) {
    // Convert the color stops to css strings
    stops = stops.map((c) => c.css || c);
    const uint8arrays = this.gradientImageData(nColors, stops, locs);
    const typedColors = this.typedArraytoColors(uint8arrays);
    Object.setPrototypeOf(typedColors, this.ColorMapProto);
    return typedColors;
  },
  // The most popular MatLab gradient, "jet":
  jetColors: [
    "rgb(0, 0, 127)",
    "rgb(0, 0, 255)",
    "rgb(0, 127, 255)",
    "rgb(0, 255, 255)",
    "rgb(127, 255, 127)",
    "rgb(255, 255, 0)",
    "rgb(255, 127, 0)",
    "rgb(255, 0, 0)",
    "rgb(127, 0, 0)",
  ],
  // Two other popular MatLab 'ramp' gradients are:
  // * One color: from black/white to color, optionally back to white/black.
  // stops = ['black', 'red'] or ['white', 'orange', 'black']
  // The NetLogo map is a concatenation of 14 of these.
  // * Two colors: stops = ['red', 'orange'] (blends the tow, center is white)

  // The 16 unique [CSS Color Names](https://goo.gl/sxo36X), case insensitive.
  // In CSS 2.1, the color 'orange' was added to the 16 colors as a 17th color
  // Aqua == Cyan and Fuchsia == Magenta, 19 total color names.
  // These sorted by hue/saturation/light, hue in 0-300 degrees.
  // See [Mozilla Color Docs](https://goo.gl/tolSnS) for *lots* more!

  basicColorNames:
    "white silver gray black red maroon yellow orange olive lime green cyan teal blue navy magenta purple".split(
      " "
    ),
  brightColorNames:
    "white silver red maroon yellow orange olive lime green cyan teal blue navy magenta purple".split(
      " "
    ),
  // Create a named colors colormap
  cssColorMap(cssArray, createNameIndex = false) {
    const array = cssArray.map((str) => Color.cssToUint8Array(str));
    const map = this.basicColorMap(array);
    map.cssNames = cssArray;
    // REMIND: kinda tacky? Maybe map.name.yellow? Maybe generalize for other
    // map types: map.closest(name)
    if (createNameIndex) {
      cssArray.forEach((name, ix) => {
        map[name] = map[ix];
      });
      if (map.cyan) map.aqua = map.cyan;
      if (map.magenta) map.fuchsia = map.magenta;
    }
    return map;
  },

  // Clone a colorMap. Useful if you want to mutate an existing shared map
  cloneColorMap(colorMap) {
    const keys = Object.keys(colorMap);
    const clone = this.basicColorMap(colorMap);
    forLoop(keys, (val, i) => {
      if (clone[i] === undefined) clone[val] = colorMap[val];
    });
    return clone;
  },

  // ### Shared Global ColorMaps
  // NOTE: Do NOT modify one of these, they are shared and would
  // surprise anyone useing them. Use cloneColorMap() to have your own
  // private one, or call any of the map factories above .. see below.

  // The shared global colormaps are lazy evaluated to minimize memory use.
  // NOTE: these are shared, so any change in them are seen by all users!
  LazyMap(name, map) {
    Object.defineProperty(this, name, { value: map, enumerable: true });
    return map;
  },
  get Gray() {
    return this.LazyMap("Gray", this.grayColorMap());
  },
  get LightGray() {
    return this.LazyMap("LightGray", this.grayColorMap(200));
  },
  get DarkGray() {
    return this.LazyMap("DarkGray", this.grayColorMap(0, 100));
  },
  get Jet() {
    return this.LazyMap("Jet", this.gradientColorMap(256, this.jetColors));
  },
  get Rgb256() {
    return this.LazyMap("Rgb256", this.rgbColorCube(8, 8, 4));
  },
  get Rgb() {
    return this.LazyMap("Rgb", this.rgbColorCube(16));
  },
  get Basic16() {
    // 17 unique + 2 "aliases" = 19 names. "16" historic
    return this.LazyMap(
      "Basic16",
      this.cssColorMap(this.basicColorNames, true)
    );
  },
  get Bright16() {
    // Basic16 w/o grays: white, black
    return this.LazyMap(
      "Bright16",
      this.cssColorMap(this.brightColorNames, true)
    );
  },
};

/**
 * Basic 2D view.
 *
 * **TODO: Document this class**
 */
class TwoDraw extends TwoView {
  static defaultOptions() {
    return {
      patchesColor: "random",
      initPatches: null,

      turtlesColor: "random",
      turtlesStrokeColor: "random",
      turtlesShape: "dart",
      turtlesSize: 1,

      linksColor: "random",
      linksWidth: 1,

      textProperty: null,
      textSize: 0.5,
      textColor: "black",

      // patchesMap: ColorMap.DarkGray,
      // turtlesMap: ColorMap.Basic16,
      patchesMap: "DarkGray",
      turtlesMap: "Basic16",
    };
  }

  // ======================

  constructor(model, viewOptions = {}, drawOptions = {}) {
    // drawOptions = Object.assign({}, TwoDraw.defaultOptions(), drawOptions)
    if (viewOptions.drawOptions) {
      drawOptions = viewOptions.drawOptions;
      delete viewOptions.drawOptions;
    }
    drawOptions = Object.assign(TwoDraw.defaultOptions(), drawOptions);

    if (typeof drawOptions.turtlesMap === "string")
      drawOptions.turtlesMap = ColorMap[drawOptions.turtlesMap];
    if (typeof drawOptions.patchesMap === "string")
      drawOptions.patchesMap = ColorMap[drawOptions.patchesMap];

    super(model.world, viewOptions);
    this.model = model;
    this.view = this;
    this.checkParams(drawOptions);
    this.drawOptions = drawOptions;
  }

  // name need not end ".png", will be added if needed.
  // if name undefined use model name, lowerCase'd w/ "Model" removed
  // downloadCanvas(name = undefined) {
  //     if (!name)
  //         name = this.model.constructor.name
  //             .toLowerCase()
  //             .replace(/model$/, '')
  //     util.downloadCanvas(this.canvas, name)
  // }

  // The parameters are easily mistaken: check they are all in the defaults.
  checkParams(params) {
    const keys = Object.keys(params);
    const defaults = TwoDraw.defaultOptions();
    keys.forEach((k) => {
      if (defaults[k] === undefined) {
        console.log(
          "Legal TwoDraw parameters",
          Object.keys(TwoDraw.defaultOptions())
        );
        throw Error("Unknown TwoDraw parameter: " + k);
      }
    });
  }

  draw() {
    // params = Object.assign({}, TwoDraw.defaultOptions(), params)
    let {
      patchesColor,
      initPatches,

      turtlesColor,
      turtlesStrokeColor,
      turtlesShape,
      turtlesSize,

      linksColor,
      linksWidth,

      textProperty,
      textSize,
      textColor,

      patchesMap,
      turtlesMap,
    } = this.drawOptions;
    const { model, view } = this;

    if (view.ticks === 0) {
      if (textProperty) view.setTextProperties(textSize);

      if (initPatches) {
        // colors is an array of typedColors or pixels:
        const colors = initPatches(model, view);
        view.createPatchPixels((i) => colors[i]);
      } else if (patchesColor === "random") {
        // NOTE: random colors only done once for patches.
        view.createPatchPixels((i) => patchesMap.randomColor());
      }
    }

    if (patchesColor === "random" || initPatches) {
      view.drawPatches(); // redraw cached patches colors below
    } else if (typeof patchesColor === "function") {
      view.drawPatches(model.patches, (p) => patchesColor(p));
    } else if (isImageable(patchesColor)) {
      view.drawPatchesImage(patchesColor);
    } else {
      view.clear(patchesColor);
    }

    const checkColor = (agent, color) =>
      color === "random" ? turtlesMap.atIndex(agent.id).css : color;
    // const checkColor = (agent, color) => {
    //     if (!turtlesMap) debugger
    //     color === 'random' ? turtlesMap.atIndex(agent.id).css : color
    // }

    view.drawLinks(model.links, (l) => ({
      color:
        linksColor === "random"
          ? turtlesMap.atIndex(l.id)
          : typeof linksColor === "function"
          ? checkColor(l, linksColor(l))
          : linksColor,
      width: linksWidth,
    }));

    view.drawTurtles(model.turtles, (t) => ({
      shape:
        typeof turtlesShape === "function" ? turtlesShape(t) : turtlesShape,
      color:
        turtlesColor === "random"
          ? turtlesMap.atIndex(t.id).css
          : typeof turtlesColor === "function"
          ? checkColor(t, turtlesColor(t))
          : turtlesColor,
      strokeColor:
        turtlesStrokeColor === "random"
          ? turtlesMap.atIndex(t.id + 4).css
          : typeof turtlesColor === "function"
          ? checkColor(t, turtlesColor(t))
          : turtlesColor,
      size: typeof turtlesSize === "function" ? turtlesSize(t) : turtlesSize,
    }));

    if (textProperty) {
      model.turtles.ask((t) => {
        if (t[textProperty] != null)
          view.drawText(t[textProperty], t.x, t.y, textColor);
      });
    }

    view.tick();
  }
}

class Animator {
  constructor(fcn, steps = -1, fps = 30) {
    Object.assign(this, { fcn, steps, fps, timeoutID: null, ticks: 0 });
    this.start();
  }
  start() {
    if (this.timeoutID) return; // avoid multiple starts
    this.timeoutID = setInterval(() => this.step(), 1000 / this.fps);
    return this; // chaining off ctor
  }
  stop() {
    if (this.timeoutID) clearInterval(this.timeoutID);
    this.timeoutID = null;
    return this; // chaining off ctor
  }
  step() {
    if (this.steps === 0) return this.stop();
    this.ticks++;
    this.steps--;
    this.fcn();
    if (this.stats) this.stats.update();
    return this; // chaining off ctor
  }

  isRunning() {
    return this.timeoutID != null;
  }

  // startStats() {
  //   if (this.stats) return console.log("startStats: already running");
  //   import("https://cdn.skypack.dev/stats.js").then((m) => {
  //     this.stats = new m.default();
  //     document.body.appendChild(this.stats.dom);
  //   });
  //   return this; // chaining off ctor
  // }

  // Stop and restart with the new steps & fps
  reset(steps = -1, fps = this.fps) {
    this.stop();
    this.steps = steps;
    this.ticks = 0;
    this.fps = fps;
    this.start();
  }
  // stop if running, start otherwise
  // if starting and steps === 0, reset with steps = -1, forever.
  toggle() {
    if (this.timeoutID) this.stop();
    else if (this.steps === 0) this.reset();
    else this.start();
  }
  // call the fcn once. stops if currently running
  once() {
    this.stop();
    this.step();
  }
}

var css_248z = ":root {\n  --background: #fff;\n  --font-color: #494949;\n}\n\n@media (prefers-color-scheme: dark) {\n  :root {\n    --background: #3c3c3c;\n    --font-color: #fafafa;\n  }\n}\n.foo-bar {\n  font-family: \"Avenir Next\", Helvetica, Arial, sans-serif;\n  color: #005f20;\n}";
styleInject(css_248z);

var Agentscript = function (_a) {
    var view = _a.view, animation = _a.animation, Model = _a.Model;
    useEffect(function () {
        var agentSetup = function () { return __awaiter(void 0, void 0, void 0, function () {
            var model, width, drawOptions, twoDraw, step, fps;
            return __generator(this, function (_a) {
                switch (_a.label) {
                    case 0:
                        model = new Model();
                        return [4 /*yield*/, model.startup()];
                    case 1:
                        _a.sent();
                        model.setup();
                        width = view.width, drawOptions = view.drawOptions;
                        twoDraw = new TwoDraw(model, {
                            div: 'modelDiv',
                            useSprites: true,
                            width: width,
                            drawOptions: drawOptions,
                        });
                        step = animation.step, fps = animation.fps;
                        new Animator(function () {
                            model.step();
                            twoDraw.draw();
                        }, step, fps);
                        return [2 /*return*/];
                }
            });
        }); };
        agentSetup();
    }, []);
    return (React.createElement("div", { id: "modelDiv" }));
};

export { Agentscript, TestComponent };
//# sourceMappingURL=index.esm.js.map
